# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 1/19/20.


import argparse
import locale
import os
import re
import sys
import time
import warnings
from pytonik import Version
from typing import Any, Callable, Dict, List, Pattern, Union

import webbrowser

from random import randint

try:
    import readline

    if readline.__doc__ and 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
        USE_LIBEDIT = True
    else:
        readline.parse_and_bind("tab: complete")
        USE_LIBEDIT = False
except ImportError:
    USE_LIBEDIT = False

def __(string):
    return string


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage='%(prog)s [OPTIONS] <PROJECT_DIR>',
        epilog=__("For more information, visit < https://pytonik.readthedocs.io >."),
        description=__("""Pytonik is a python framework built to enhance web development
        fast and easy, also help web developers to build more apps with less codes.
        it uses expressive architectural pattern, structured on model view controller MVC
        and bundles of component to reuse while deploying the framework."""))

    parser.add_argument('-q', '--quit', action='store_true', dest='quit',
                        default=None,
                        help=__('quit mode'))
    parser.add_argument('--version', action='version', dest='show_version',
                        version='%%(prog)s %s' % Version.VERSION_TEXT)

    parser.add_argument('path', metavar='PROJECT_DIR', default='.', nargs='?',
                        help=__('project root'))

    return parser


def main(argv: List[str] = sys.argv[1:]) -> int:
    # parse options
    parser = get_parser()
    try:
        args = parser.parse_args(argv)

    except SystemExit as err:
        return err.code
    d = vars(args)
    import http.server
    import socketserver

    class MyHttpRequestHandler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            if self.path == '/':
                self.path = 'docs'
            if self.path.endswith('favicon.ico'):
                    return
            if self.path.endswith('robots.txt'):
                    return
            if self.path.endswith('favicon.ico'):
                    return
            return http.server.SimpleHTTPRequestHandler.do_GET(self)

    # Create an object of the above class
    handler_object = MyHttpRequestHandler

    PORT = 6061
    url = "localhost"
    #my_server = socketserver.TCPServer((url, PORT), handler_object)

    l = "https://pytonik.readthedocs.io/en/latest/" #"http://{}:{}".format(url, PORT)
    print("Documentation Link {}".format(l))
    try:
        webbrowser.open_new(l)
    except Exception as err:
        print(err)
    # Star the server
    #my_server.serve_forever()



if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
#############################################
#############################################
# MYSQL Support Database Connection
# Using Raw query from mysql-connector module
# Each method represent extensions from mysql version 2.2.9


from pytonik import Log
log_msg = Log.Log()

try:
    import mysql.connector
except Exception as err:
    log_msg.critical(err)


class MYSQL:

    def __init__(self, setting):
        self.settings = setting
        self.host = setting.get('host', '') #  get 'host' name 
        self.username = setting.get('username', '') #  get host 'username' 
        self.password = setting.get('password', '') #  get host 'password' 
        self.port = setting.get('port', '') #  get host 'port' 
        self.database = setting.get('database', '') #  get 'database' name 
        self.prefix = setting.get('prefix', '') #  get database 'prefix' 
        self.Exception = "" # Exception String handler
        self.conn =  None # Connection Return handler
        self.con = None # Connection Return cursor handler
        self.result = None # Return Query Results
        self.connectDB() # Initiating Connection Method 


    # This method 'connectDB' represent MYSQL configuration and Connection using 
    # 'host', 'database' 'username',  'password',  'prefix' 
    def connectDB(self):

        try:

            self.conn = mysql.connector.connect(
                    host=self.host,
                    user=self.username,
                    passwd=self.password,
                    database=self.database,
                    port = self.port
            )
        except mysql.connector.Error as err:
            log_msg.error(err)
            self.Exception = err

    # This method 'query' execute MYSQL query and returns cursor handler for 
    # select, update, delete, insert, tasks 
    def query(self, sql="", value = ""):
        try:
            self.con = self.conn.cursor(dictionary=True)
            if sql !="" and value != "":
                self.con.execute(str(sql), value)
            else:
                self.con.execute(str(sql))
        except Exception as err:
            log_msg.error(err)
            self.Exception = err
            
        return self

    # This method 'querymultiple' handles many execution of MYSQL queries and returns cursor handler for 
    # multiple insert tasks 
    def querymultiple(self, sql="", value = ""):
        try:
            self.con = self.conn.cursor(dictionary=True)
            
            if sql !="" and value != "":
                self.con.executemany(str(sql), value)
            else:
                self.con.executemany(str(sql))
        except Exception as err:
            log_msg.error(err)
            self.Exception = err
            
        return self

    # This method 'lastId' gets the last inserted or generated increment number in a row
    def lastId(self):
        return self.con.lastrowid

    # This method 'fetch' returns table results dictionary
    def fetch(self):
            return self.con.fetchall()

    # This method 'queryone' execute MYSQL query and returns cursor handler for 
    # select, update, delete, insert, tasks 
    def queryone(self, sql="", value = ""):
        self.con = self.conn.cursor(buffered=True)
        if sql !="" and value != "":
            self.con.execute(str(sql), value)
        else:
            self.con.execute(str(sql))
            
        return self.con

    # This method 'all' returns table results dictionary
    # Helps in calling both result and count
    def all(self):
        self.result = self.fetch()
        return self.result

    # This method 'one' returns table results list
    def one(self):
        return self.con.fetchone()

    # This method 'count' returns total number of rows count of a table
    def count(self):
        return self.con.rowcount
    
    # This method 'countall' returns table results dictionary
    # Helps in calling both result and count
    def countall(self):
        self.all()
        return self.con.rowcount

    # This method 'save' Execute table query: 
    # insert, update and return bool -> True or Exception
    def save(self):
        try:
            self.conn.commit()
            return True
        except Exception as err:
            self.Exception = err
            log_msg.error(err)
            return self

    # This method 'close' end query and also 
    # close connection at the end of a query 
    def close(self):
        return self.con.close()

    # This method 'create' handles query creation of tables
    def create(self, TABLES = ''):
        self.con = self.conn.cursor()
        if TABLES:
            for table_name in TABLES:
                table_description = TABLES[table_name]
                try:
                    self.con.execute(table_description)
                    
                except mysql.connector.Error as err:
                        log_msg.info("Database table '{}' already exists.".format(table_name))
                        self.Exception =  "Database table '{}' already exists.".format(table_name)
            return self          
        else:
            self.Exception =  "Empty Table"
            return self



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

import sys, os
from pytonik import Version, Log
from pytonik.Config import Config
from pytonik.Core.env import env
from pytonik.Session import Session
from pytonik.util.Variable import Variable
from .Core import Helpers
h = Helpers

class Controllers(env, Config):

    def __getattr__(self, item):
        return item


    def __init__(self):
        osv = Variable()
        url = osv.out('REQUEST_URI', "")
        http_s = osv.out("HTTP_HOST")

        if osv.out("SERVER_SOFTWARE", "") == Version.AUTHOR:

            self.uri = url.split('/')[2:]

        else:

            if http_s == "127.0.0.1" or http_s == "localhost":
                self.uri = url.split('/')[2:]
            else:
                self.uri = url.split('/')[1:]

        self.add(self._e())

        self.controllers = self.get('default_controllers')
        self.actions = self.get('default_actions')
        self.default_controllers = self.get('default_controllers')
        self.default_actions = self.get('default_actions')

        self.languages = self.get('default_languages')
        self.all_languages = self.get('languages', '')
        self.routes = self.get('default_routes', "")
        self.methodprefix = ""
        self.parameter = ""

        path_parts = self.uri

        # Get parameter from Query String
        pathparts_paramarray = os.environ.get("QUERY_STRING", '')

        pathparts_paramarrayOut = dict()
        if pathparts_paramarray != '':
            pairs = pathparts_paramarray.split('&')

            pathparts_paramarray = pairs

            for i in pairs:

                name, value = i.split('=', 2)

                pathparts_paramarray = {name: value}

                if name in pathparts_paramarray:

                    pathparts_paramarray[name] = value

                else:
                    pathparts_paramarray[name] = [pathparts_paramarray[name], value]

                pathparts_paramarrayOut.setdefault(name, value)


        else:
            pathparts_paramarrayOut = ""


        if len(path_parts):


            if Version.PYVERSION_MA < 3:
                path_parts = filter(None, path_parts)
            else:
                path_parts = list(filter(None, path_parts))

            routes = self.get('route', '')

            if list(set(path_parts).intersection(routes.keys())):

                for s in path_parts:

                    if s in routes:
                        self.routes = s

                        if self.routes in routes:
                            self.methodprefix = routes[self.routes]
                        else:
                            self.methodprefix = ""

                            # path_parts.append(path_parts.pop(-1))


            if list(set(path_parts).intersection(self.all_languages.keys())):

                for s in path_parts:
                    if s in self.all_languages:
                        self.languages = s
                        path_parts.append(path_parts.pop(-1))

            controllers = self.get('default_controllers', '')
            if controllers:

                i = 0
                path_parts = list(filter(None, path_parts))

                for s in path_parts:

                    if s is not self.languages:
                        i += 1

                        if i == 1:
                            self.controllers = s
                            path_parts.append(path_parts.pop(-1))
                    ++i

            action = self.get('default_actions', '')
            if action:
                i = 0
                for s in path_parts:
                    if s is not self.controllers and s is not self.languages:
                        i += 1
                        if i == 1:
                            self.actions = s
                            path_parts.append(path_parts.pop(-1))
                        ++i


            # Get Path from URI / convert it to parameter
            list_params = []

            if pathparts_paramarray == None or pathparts_paramarray == "":
                if Version.PYVERSION_MA <= 2:
                    lroutes = routes.iteritems()
                else:
                    lroutes = routes.items()
                for k, getRouter in lroutes:

                    if self.controllers == k:

                        paraUri = getRouter.split('@')
                    else:
                        paraUri = []
                    if len(paraUri) > 0:
                        if ':' not in paraUri[1]:
                            getMapPara = []
                        else:
                            getMapPara = paraUri[1].split(':')

                        if self.controllers in routes:

                            if len(getMapPara[1:]) > 0:

                                new_para = path_parts[1:]

                                if len(new_para) > 0:
                                    param_m = []

                                    for i, para in enumerate(getMapPara[1:]):

                                        param_n = para

                                        if (len(new_para) - i) > 0:
                                            v_para = new_para[i]
                                        else:
                                            v_para = ""

                                        list_params.append(param_n)
                                        list_params.append(v_para)

                                self.parameter = Helpers.covert_list_dict(list_params)


                else:
                    for s in path_parts:

                        if s is not self.controllers and s is not self.actions and s is not self.languages:
                            list_params.append(s)

                            path_parts.append(path_parts.pop(-1))

                    self.parameter = Helpers.covert_list_dict(list_params)

            else:

                self.parameter = pathparts_paramarrayOut

                path_parts.append(path_parts.pop(-1))

        return None

    def _getUri(self):
        return self.uri

    def _getControllers(self):
        return self.controllers

    def _getActions(self):
        return self.actions

    def _getLanguages(self):
        return self.languages

    def _getParams(self):
        return self.parameter

    def _getMethodPrefix(self):

        return self.methodprefix

    def _getRoutes(self):
        return self.routes


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###



import re, operator, ast
from pytonik import Version, App
import os, importlib, sys
from pytonik.Log import Log

VAR_FRAGMENT = 0
OPEN_BLOCK_FRAGMENT = 1
CLOSE_BLOCK_FRAGMENT = 2
TEXT_FRAGMENT = 3
CLOSE_COMMENT_FRAGMENT = 4

VAR_TOKEN_START = '{{'
VAR_TOKEN_END = '}}'
BLOCK_TOKEN_START = '{%'
BLOCK_TOKEN_END = '%}'
COMMENT_TOKEN_START = '{#'
COMMENT_TOKEN_END = '#}'



TOK_REGEX = (re.compile('(%s.*?%s|%s.*?%s|%s.*?%s)' % (
    re.escape(VAR_TOKEN_START),
    re.escape(VAR_TOKEN_END),
    re.escape(BLOCK_TOKEN_START),
    re.escape(BLOCK_TOKEN_END),
    re.escape(COMMENT_TOKEN_START),
    re.escape(COMMENT_TOKEN_END)
)))

TRANSLATOR_COMMENT_MARK = 'Comment'

WHITESPACE = re.compile("[\s]")
UPPARA = re.compile('\,+')

operator_lookup_table = {
    '<': operator.lt,
    '>': operator.gt,
    '==': operator.eq,
    '!=': operator.ne,
    '<=': operator.le,
    '>=': operator.ge,
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '//': operator.floordiv,
    '/': operator.truediv,
    '%': operator.mod,
    '**': operator.pow,
    '<<': operator.lshift,
    '>>': operator.rshift,
    '^': operator.xor,
}

if os.path.isdir(os.getcwd() + '/public'):
    host = os.getcwd()  # os.path.dirname(os.getcwd())

else:
    host = os.path.dirname(os.getcwd())


class TemplateError(Exception):
    pass


class TemplateContextError(TemplateError):

    def __init__(self, context_var):
        self.context_var = context_var


    def __str__(self):
        Log('').error("cannot resolve '%s'" % self.context_var)
        return "cannot resolve '%s'" % self.context_var


class TemplateSyntaxError(TemplateError):

    def __init__(self, error_syntax):
        self.error_syntax = error_syntax

    def __str__(self):
        Log('').error("'%s' seems like invalid syntax" % self.error_syntax)
        return "'%s' seems like invalid syntax" % self.error_syntax


def eval_expression(expr):

    try:
        return 'literal', ast.literal_eval(expr)
    except Exception as err:
        return 'name', expr


def resolve(name, context):

    if name.startswith('..'):
        context = context.get('..', {})
        name = name[2:]

    try:

        for tok in name.split('.'):
            context = context[tok]
        return context


    except Exception as err:


        try:
            Ap = App.App()
            load = Ap.loadmodule()
            load.update({name: name})
            md = importlib.import_module(name, name)
            ob = getattr(md, name)
            return ob()
        except Exception as err:

            # name
            raise TemplateContextError(err)


class _Fragment(object):
    def __init__(self, raw_text):

        self.raw = raw_text

        self.clean = self.clean_fragment()

    def clean_fragment(self):
        if self.raw[:2] in (VAR_TOKEN_START, BLOCK_TOKEN_START):

            return self.raw.strip()[2:-2].strip()

        return self.raw

    @property
    def type(self):
        raw_start = self.raw[:2]
        if raw_start == VAR_TOKEN_START:
            return VAR_FRAGMENT
        elif raw_start == BLOCK_TOKEN_START:

            block = str('end{}'.format(self.clean[3:9]))

            return CLOSE_BLOCK_FRAGMENT if self.clean[:9] == block else OPEN_BLOCK_FRAGMENT

        elif raw_start == COMMENT_TOKEN_START:
            if self.clean.find(TRANSLATOR_COMMENT_MARK):
                self.clean[2:-2].strip()
            return CLOSE_COMMENT_FRAGMENT
        else:

            return TEXT_FRAGMENT



class _Node(object):
    creates_scope = False

    def __init__(self, fragment=None):
        self.children = []

        self.process_fragment(fragment)

    def process_fragment(self, fragment):

        pass

    def enter_scope(self):
        pass

    def render(self, context):

        pass

    def exit_scope(self):
        pass

    def render_children(self, context, children=None):


        if children is None:
            children = self.children

        def render_child(child):

            child_html = child.render(context)

            return '' if not child_html else str(child_html)

        return ''.join(map(render_child, children))


class _ScopableNode(_Node):
    creates_scope = True


class _Root(_Node):
    def render(self, context):

        return self.render_children(context)


class _Variable(_Node):
    def process_fragment(self, fragment):
        self.name = fragment

    def render(self, context):

        return resolve(self.name, context)


class _Each(_ScopableNode):

    def process_fragment(self, fragment):

        try:

            _, it = WHITESPACE.split(fragment, 1)
            self.it = eval_expression(it)

        except Exception as err:

            raise TemplateSyntaxError(fragment)

    def render(self, context):

        items = self.it[1] if self.it[0] == 'literal' else resolve(self.it[1], context)

        def render_item(item):
            Ap = App.App()
            load = Ap.loadmodule()

            load_m = {'..': context, 'it': item}

            load.update(load_m)

            return self.render_children(load)

        return ''.join(map(render_item, items))

class _Block(_Node):

    def process_fragment(self, fragment):

        try:
            it = WHITESPACE.split(fragment)

            self.block = eval_expression(it)

        except Exception as err:
            raise TemplateSyntaxError(fragment)

    def render(self, context):

        it = self.block[1]

        d, s = [], []
        try:

            for iv in it:
                if iv not in operator_lookup_table:
                    items = resolve(iv, context)
                    d.append(items)
                else:
                    s.append(iv)
            operator_lookup = s[0]
            try:
                return eval(operator_lookup.join(d))

            except Exception as err:
                return operator_lookup.join(d)

        except Exception as err:

            for iv in it:

                if iv not in operator_lookup_table:
                    d.append(iv)
                else:
                    s.append(iv)

            operator_lookup = s[0]
            try:
                return eval(operator_lookup.join(d))
            except Exception as err:
                return operator_lookup.join(d)





class _If(_ScopableNode):
    def process_fragment(self, fragment):
        bits = fragment.split()[1:]
        if len(bits) not in (1, 3):
            raise TemplateSyntaxError(fragment)
        self.lhs = eval_expression(bits[0])
        if len(bits) == 3:
            self.op = bits[1]
            self.rhs = eval_expression(bits[2])

    def render(self, context):

        lhs = self.resolve_side(self.lhs, context)

        if hasattr(self, 'op'):
            op = operator_lookup_table.get(self.op)
            if op is None:
                raise TemplateSyntaxError(self.op)
            rhs = self.resolve_side(self.rhs, context)
            exec_if_branch = op(lhs, rhs)
        else:
            exec_if_branch = operator.truth(lhs)


        self.if_branch, self.else_branch = self.split_children()
        return self.render_children(context, self.if_branch if exec_if_branch else self.else_branch)

    def resolve_side(self, side, context):

        return side[1] if side[0] == 'literal' else resolve(side[1], context)

    def exit_scope(self):
        self.if_branch, self.else_branch = self.split_children()

    def split_children(self):
        if_branch, else_branch = [], []
        curr = if_branch
        for child in self.children:
            if isinstance(child, _Else):
                curr = else_branch
                continue
            curr.append(child)
        return if_branch, else_branch


class _Else(_Node):
    def render(self, context):
        pass


def dict_local(it, resolves):

    for i, k in enumerate(resolves):
        if k in resolves:

            l = str(it).replace(k, str(resolves[k]))
        if len(resolves) > 1:
            return dict_local_next(l, resolves)
        else:
            return l

def dict_local_next(it, resolves):

    for i, k in enumerate(resolves):

        if k in it:
           l = str(it).replace(k, str(resolves[k]))
           return l

class _Call(_Node):

    def process_fragment(self, fragment):

        try:
            #bits = WHITESPACE.
            self.bits = WHITESPACE.split(fragment)
            self.callable = self.bits[1]
            self.fragment = fragment

            self.args, self.kwargs = self._parse_params(self.bits[2:])



        except Exception as err:
            raise TemplateSyntaxError(fragment)

    def _parse_params(self, params):
        new_args, args, kwargs = [], [], {}

        for param in params:

            if len(params) > 3:

                p = str(params).translate({ord(i): None for i in '"'})

                if '=' in p:

                    new_name = str(p).translate({ord(i): None for i in '"\',\[]'})
                    sln = new_name.split(" ")

                    for n in sln:

                        if '=' in n:

                            k, v = str(n).split("=")

                            kwargs[k] = eval_expression(str(v))
                        else:

                            new_n = str(n).translate({ord(i): None for i in '[]'})
                            new_args.append(new_n)


                    args.append(eval_expression(" ".join(new_args)))

                else:

                    args.append(eval_expression(p))
                break

            else:
                if '=' in param:
                    name, value = param.split('=')

                    kwargs[name] = eval_expression(str(value))
                else:
                    args.append(eval_expression(param))


        return args, kwargs

    def render(self, context):
        self.contxt = context

        ob_dir = [str(os.path.dirname(__file__).replace('Editor', '')) + str("Functions"), str(host) + str("/") + "model"]



        resolved_args, resolved_kwargs = [], {}


        for kind, value in self.args:

            if kind == 'name':
                value = value
            value = self._call_each(str(value))
            resolved_args.append(value)


        if Version.PYVERSION_MA >= 2:
            items  = self.kwargs.items()
        else:
            items = self.kwargs.iteritems()

        for key, (kind, value) in items:

            if kind == 'name':

                value = value

            value = self._call_each(str(value))
            try:
                valux = eval(value)
            except Exception as err:
                valux = value
            resolved_kwargs[key] = valux



        path = [str(ob_dir[0]) + "/" + str(self.callable) + ".py",  str(ob_dir[1]) + "/" + str(self.callable) + ".py"]


        sys.path.append(str(ob_dir[0]))
        sys.path.append(str(ob_dir[1]))


        importlib._RELOADING

        if os.path.isfile(path[0]) == True:


            md = importlib.import_module(self.callable, self.callable)


            ob = getattr(md, self.callable)



            if hasattr(ob(), '__call__') == True:

                calls = ""
                try:
                    _cal = ob()
                    _new_cal = getattr(_cal, *resolved_args)
                    calls =  _new_cal(**resolved_kwargs)
                except Exception as err:

                    try:
                        newob = getattr(ob, ''.join(resolved_args))

                        calls = newob(**resolved_kwargs)

                    except Exception as err:
                        try:
                            calls = ob(*resolved_args, **resolved_kwargs)

                        except Exception as err:
                            Log('').error(err)

                return calls
            else:
                raise TemplateError("'%s' is not a callable" % self.callable)

        elif os.path.isfile(path[1]) == True:


            md = importlib.import_module(self.callable, self.callable)
            ob = getattr(md, self.callable)


            if hasattr(ob(), '__call__'):

                calls = ""
                try:
                    _cal = ob()
                    _new_cal = getattr(_cal, *resolved_args)
                    calls = _new_cal(**resolved_kwargs)
                except Exception as err:

                    try:
                        newob = getattr(ob, ''.join(resolved_args))

                        calls = newob(**resolved_kwargs)

                    except Exception as err:
                        try:
                            calls = ob(*resolved_args, **resolved_kwargs)

                        except Exception as err:
                            Log('').error(err)
                return calls
            else:
                raise TemplateError("'%s' is not a callable" % self.callable)

        else:
            raise TemplateError("'%s' module not found " % self.callable)

    def _call_each(self, context):

        if  VAR_TOKEN_START in context:
            self.it = str(context).translate({ord(i): None for i in '{VAR_TOKEN_START}{VAR_TOKEN_END}'.format(VAR_TOKEN_START = '{{', VAR_TOKEN_END = '}}')})
            oparatork = "/"
            for oparator_k,  oparator_v in operator_lookup_table.items():
                if oparator_k in self.it:
                    oparatork = oparator_k


            contsplit = self.it.split(oparatork)


            if len(contsplit) < 2:

                for k, itc in enumerate(contsplit):

                    if 'it.' in itc:

                        ite = resolve(itc, self.contxt)
                        item = str(self.it).replace(itc, str(ite))
                        return item
                    elif '..' in itc:

                        ite = resolve(itc, self.contxt)
                        item = str(self.it).replace(itc, str(ite))
                        return item
                    else:
                        ite = resolve(itc, self.contxt)
                        item = str(self.it).replace(itc, str(ite))
                        return item


            elif len(contsplit) > 1:

                dic_ls = {}
                for s in contsplit:

                    if 'it.'.lower() in s.lower():
                        dic_ls.update({s:resolve(s, self.contxt)})

                    elif '..'.lower() in s.lower():

                        dic_ls.update({s: resolve(s, self.contxt)})



                return dict_local(self.it,dic_ls)


            else:

                if 'it.' in self.it:
                    ite = resolve(self.it, self.contxt)
                    item = str(self.it).replace(self.it, ite)
                elif '..' in self.it:

                    ite = resolve(self.it, self.contxt)
                    item = str(self.it).replace(self.it, ite)

                return item



        elif '..' in context:

            for it in context.split('/'):
                if '..' in it:
                    ite = resolve(it, self.contxt)

                else:
                    ite = resolve(it, self.contxt)

                item = str(context).replace(it, ite)
                return item

        else:

            return context



class _Text(_Node):
    def process_fragment(self, fragment):
        self.text = fragment

    def render(self, context):

        return self.text


class Compiler(object):
    def __init__(self, template_string):
        self.template_string = template_string

    def each_fragment(self):


        for fragment in TOK_REGEX.split(self.template_string):

            if fragment:
                yield _Fragment(fragment)

    def compile(self):

        root = _Root()
        scope_stack = [root]

        for fragment in self.each_fragment():

            if not scope_stack:
                raise TemplateError('nesting issues')

            parent_scope = scope_stack[-1]


            if fragment.type == CLOSE_BLOCK_FRAGMENT:
                parent_scope.exit_scope()
                scope_stack.pop()
                continue

            new_node = self.create_node(fragment)

            if new_node:

                parent_scope.children.append(new_node)
                if new_node.creates_scope:
                    scope_stack.append(new_node)
                    new_node.enter_scope()

        return root

    def create_node(self, fragment):
        node_class = None
        if fragment.type == TEXT_FRAGMENT:
            node_class = _Text
        elif fragment.type == VAR_FRAGMENT:
            node_class = _Variable
        elif fragment.type == OPEN_BLOCK_FRAGMENT:
            cmd = fragment.clean.split()[0]

            if cmd == 'each':
                node_class = _Each
            elif cmd == 'if':
                node_class = _If
            elif cmd == 'else':
                node_class = _Else
            elif cmd == 'call':
                node_class = _Call
            else:
                node_class = _Block

        if node_class is None:
            raise TemplateSyntaxError(fragment)

        return node_class(fragment.clean)


class Template(object):
    def __init__(self, contents):
        self.contents = contents
        self.root = Compiler(contents).compile()



    def render(self, **kwargs):
        return self.root.render(kwargs)



from pytonik.Web import App, Version

app = App()


def index():
    data = {
        'title': 'Welcome',
        'name': 'Pytonik',
        'version': Version().VERSION_TEXT,
        'reversion': Version().VERSION_TEXT[0:5],
    }

    return app.views('index', data)



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


from pytonik.Request import Request
from pytonik.Session import Session
from pytonik.Editor import HTMLeditor
from pytonik.Config import Config
from pytonik.Log import Log
from pytonik import Lang
from pytonik import Version
from pytonik.Core.env import env
from pytonik.util.Variable import Variable
from pytonik.Functions import url
from pytonik.Controllers import Controllers
import os, sys, cgi, cgitb, importlib, glob, inspect

cgitb.enable()

global host, u

if os.path.isdir(os.getcwd() + '/public'):
    host = os.getcwd()  # os.path.dirname(os.getcwd())

else:
    host = os.path.dirname(os.getcwd())

DS = str("/")
u = url

controllerpath = host + DS + 'controller'
error_page_class = controllerpath + DS + "ErrorController" + ".py"
header_response_page = {
    '400': 'error/page400',  # Bad Request
    '404': 'error/page404',  # Not Found
    '405': 'error/page405',  # Method Not Allowed
    '500': 'error/',  # Internal Server Error

}


class App(env, Config, Variable):
    def __getattr__(self, item):
        return item

    def __init__(self, routers="", routes="", getpath="", getrouters=""):

        self.routers = routers
        self.routes = routes
        self.getpath = getpath
        self.codereponse = 200
        self.getrouters = getrouters
        self.getDB = ""
        self.Request = Request
        self.Session = Session
        self.methodprefix = ""
        self.actions = ""
        self.controllers = ""
        self.controller = ""
        self.dbDriver = None
        self.Config = None
        self.url = ""
        self.params = ""
        self.key = ""
        self.vpathf = ""
        self.languages = ""
        self.Driver = ""
        self.formData = None

    def getRouters(self):

        return self.routers

    def getPath(self):
        return self.getpath

    def runs(self, formData=None):

        self.formData = formData
        return self.initial()

    def initial(self):
        self.default("Framework", Version.AUTHOR)
        self.default("X-Version", Version.VERSION_TEXT)
        self.getrouters = self.envrin('route')
        self.controllers = Controllers()
        self.methodprefix = self.controllers._getMethodPrefix()
        self.actions = self.controllers._getActions()
        self.uri = self.controllers._getUri()
        self.controller = self.controllers._getControllers()
        self.routers = self.controllers._getRoutes()
        self.languages = self.controllers._getLanguages()
        self.params = self.controllers._getParams()

        langs = Lang.Lang(self.languages)
        langs.loadLang()
        controlUri = []

        if Version.PYVERSION_MA >= 3:
            lrounters = self.getrouters.items()
        else:
            lrounters = self.getrouters.iteritems()

        for k, getRouter in lrounters:

            if self.controller == k:
                controlUri = getRouter.split('@')

        if os.path.isfile(host + DS + "routes.py") == True:
            sys.path.append(host)
            # from pytonik.Router import Router
            # route = Router()
            import routes as route

            if len(route.route.getRouter()) > 0:

                for i, redirect in enumerate(route.route.getCode()):
                    if len(route.route.getRediret()) > 0:
                        toroute = route.route.getRediret()[i]
                    if len(route.route.getDespatch()) > 0:
                        fromroute = route.route.getDespatch()[i]
                    if fromroute == self.controller:
                        self.redirect(location=toroute, link=True, code=redirect)


                for i, route_c in enumerate(route.route.getRouter()):

                    if self.controller == route_c:

                        if len(route.route.getController()) > 0:

                            self.controller = route.route.getController()[i]

                        if len(route.route.getAction()) > 0:

                            self.actions = route.route.getAction()[i]

                        if len(route.route.getRouter()) > 0:

                            self.routers = route.route.getRouter()[i]
                        if len(route.route.getMethod()) > 0:
                            self.method = route.route.getMethod()[i]

                            if self.method !="":
                                if self.method != self.out("REQUEST_METHOD", ""):

                                    Log().error(str(str(self.controller).capitalize()) + "Controller/" + str(self.actions) + " Requires " + self.method)

                                    return self.errorP('400')

        if len(controlUri) != 0:

            if 'controller' in controlUri[0].lower():

                controllersClass = str(controlUri[0].lower()).replace('controller', '').capitalize() + 'Controller'
            else:
                controllersClass = str(controlUri[0]).capitalize() + 'Controller'

            if ':' not in controlUri[1]:
                controllersMethods = str(controlUri[1])

            else:
                getMapPara = controlUri[1].split(':')

                controllersMethods = str(getMapPara[0])



        else:

            if str(self.controller[0]) == '?':

                controllersClass = 'IndexController'
            else:

                controllersClass = str(self.controller[0].capitalize()) + str(self.controller[1:]) + 'Controller'

            controllersMethods = str(self.actions)

        if controllersMethods != "":
            self.actions = controllersMethods

        controllers = controllerpath + DS + controllersClass + ".py"

        if os.path.isfile(controllers) == True:

            if __name__ == '__main__':
                spac = ""
            if os.path.isfile(controllers) == True:

                if sys.version_info.major <= 2:

                    return self.strClass(controllerpath, controllersClass)
                else:

                    return self.strClass3(controllerpath, controllersClass)

        else:
            Log(controllerpath).error('Controller does not exist ' + str(controllersClass))
            return self.errorP('405')

    def errorP(self, code, replace=""):
        getErrorP = self.envrin('error')

        pageCode = "page{code}".format(code=code)
        if os.path.isdir(os.getcwd() + '/public'):
            if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
                re_url = ""
                errorP = {}
                if getErrorP != '':
                    errorP = getErrorP

                if errorP.get(code, '') != '':

                    controllerP = ""
                    if '/' in errorP.get(code, ''):
                        splitP = errorP.get(code, '').split('/')
                        controllerP = controllerpath + DS + str(splitP[0]).capitalize() + 'Controller' + ".py"
                    else:
                        controllerP = controllerpath + DS + str(
                            errorP.get(code, '')).capitalize() + 'Controller' + ".py"

                    if code == "500":

                        if os.path.isfile(controllerP) == True:
                            re_url = u.url().url('/' + str(errorP.get(code, '')))
                        else:
                            re_url = u.url().url("/error")
                    else:
                        if os.path.isfile(controllerP) == True:
                            re_url = u.url().url('/' + str(errorP.get(code, '')))
                        else:

                            if os.path.isfile(self.error_page_html(code)) == True:
                                re_url = u.url().url("/error/{code}".format(code=pageCode))

                else:
                    if code == "500":
                        re_url = u.url().url("/error")
                    else:

                        if os.path.isfile(self.error_page_html(code)) == True:
                            re_url = u.url().url("/error/{code}".format(code=pageCode))

                return code, re_url

        if getErrorP != '':

            errorP = getErrorP

            if errorP.get(code, '') != '':
                if '/' in errorP.get(code, ''):
                    splitP = errorP.get(code, '').split('/')
                    controllerP = controllerpath + DS + str(splitP[0]).capitalize() + 'Controller' + ".py"
                else:
                    controllerP = controllerpath + DS + str(errorP.get(code, '')).capitalize() + 'Controller' + ".py"

                if os.path.isfile(controllerP) == True:
                    return self.redirect(u.url().url('/' + str(errorP.get(code, ''))))
                else:

                    if os.path.isfile(self.error_page_html(code)) == True:
                        return self.redirect(u.url().url("/error/{code}".format(code=pageCode)))

            else:

                if os.path.isfile(self.error_page_html(code)) == True:
                    return self.redirect(u.url().url("/error/{code}".format(code=pageCode)))


        else:
            if os.path.isfile(error_page_class) == True:
                return self.redirect(u.url().url("/error/{code}".format(code=pageCode)))

    def envrin(self, key):
        env = self._e()
        self.add(env)
        return self.get(key, '')

    def DB(self):

        self.dbDriver = self.envrin('dbConnect')
        if self.dbDriver.get("driver", '') == "MYSQL":
            from pytonik.Driver.DB.MYSQL.MYSQL import MYSQL
            self.getDB = MYSQL(self.dbDriver)
            self.Driver = "MYSQL"
            return self.getDB

        if self.dbDriver.get("driver", '') == "SQLite":
            from pytonik.Driver.DB.SQLite.SQLite import SQLite
            self.getDB = SQLite(self.dbDriver)
            self.Driver = "SQLite"
            return self.getDB

        if self.dbDriver.get("driver", '') == "Oracle":
            from pytonik.Driver.DB.Oracle.Oracle import Oracle
            self.getDB = Oracle(self.dbDriver)
            self.Driver = "Oracle"
            return self.getDB

        if self.dbDriver.get("driver", '') == "pyPgSQL":
            from pytonik.Driver.DB.pyPgSQL.pyPgSQL import pyPgSQL
            self.getDB = pyPgSQL(self.dbDriver)
            self.Driver = "pyPgSQL"
            return self.getDB

    def strClass(self, p=None, c=None):

        try:
            sys.path.append(p)
            ms = str(self.actions)
            md = importlib.import_module(c)
            return self.strMethod(p, md, ms)

        except Exception as err:
            Log(p + DS + c + '.py').critical(err)

            return self.errorP('400')

    def strMethod(self, p, c=None, m=None):
        Request = self.Request(prform=self.formData)
        Session = self.Session()
        try:
            return getattr(c, m)(Request, Session)
        except Exception as err:
            try:
                return getattr(c, m)(Session, Request)
            except Exception as err:
                try:
                    return getattr(c, m)(Request)
                except Exception as err:
                    try:
                        return getattr(c, m)(Session)
                    except Exception as err:
                        try:
                            return getattr(c, m)()
                        except Exception as err:

                            Log(p + DS + c + '.py').critical(err)
                            return self.errorP('400')

    def strClass3(self, p=None, c=None):

        try:
            sys.path.append(p)
            ms = str(self.actions)
            importlib._RELOADING
            md = importlib.import_module(c, ms)
            return self.strMethod(p, md, ms)

        except Exception as err:

            Log(p + DS + c + '.py').critical(err)
            return self.errorP('400')

    def redirect(self, location='/', link=False, code="307"):
        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:

            if link == True:
                location_d = u.url().url(location)
            else:
                location_d = location
            self.put(redirect_url=location_d)
            return code, location_d
        else:

            if link == True:
                location_d = u.url().url(location)
            else:
                location_d = location

            print("Location: {location}".format(location=location_d))
            print()

    def referer(self, location='/', link=False, code="307"):
        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
            if link == True:
                location_d = u.url().url(location)
            else:
                location_d = location

            self.put(referral=self.out('HTTP_REFERER', location_d))
            return code, self.out('HTTP_REFERER', location_d)
        else:
            if link == True:
                location_d = u.url().url(location)
            else:
                location_d = location
            print("Location: {location}".format(location=self.out('HTTP_REFERER', location_d)))
            print()

    @staticmethod
    def header(p=0, type="text/html"):

        if Version.PYVERSION_MA <= 2 and Version.PYVERSION_MI <= 7:
            reload(sys)
            sys.setdefaultencoding('utf-8')
        elif Version.PYVERSION_MA == 3 and Version.PYVERSION_MI <= 3:
            import imp
            imp.reload()

        elif Version.PYVERSION_MA >= 3 and Version.PYVERSION_MI >= 4:
            import importlib
            importlib.reload(sys)
        if os.path.isdir(os.getcwd() + '/public') == False:
            print("Content-type: {type}\r\n".format(type=type))  # \r\n\r\n
            if p > 0:
                for x in range(p):
                    print("")
        else:
            return

    def XHreponse(self, dataString):
        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
            return dataString
        else:
            self.header()
            print(dataString)

    def views(self, pathf="", datag={}, datal={}):

        if pathf == "":
            pathf = self.getDefaultViewPath()

        pathfhtml = host + DS + 'views' + DS + pathf + ".html"
        html = ""
        if os.path.isfile(pathfhtml) == False:
            Log(host + DS + 'views').critical('Cannot find file {}'.format(pathf + ".html"))

            return self.errorP('404')


        else:
            if os.path.isdir(os.getcwd() + '/public'):
                # print(os.environ)
                return self.read_html(host + DS + 'views' + DS, pathf, datag)

            else:
                self.header()
                print(self.read_html(host + DS + 'views' + DS, pathf, datag))

    def read_html(self, template_dir, engine, context=[]):

        html_file_path = os.path.join(template_dir, "%s.html" % engine)

        try:
            with open(html_file_path) as html_file:
                html = html_file.read()

            return str('<!-- Pytonik -->\n') + HTMLeditor.Template(html).render(**context) + str(
                '\n<!-- Pytonik {} -->'.format(Version.VERSION_TEXT))

        except Exception as err:

            Log(template_dir + DS + engine + str('.html')).error(err)
            return

    def getDefaultViewPath(self):

        router = self.Router.getRoutes()  # Routers.getRoutes()

        if router == "":
            rout = ""
        controllerDirectory = self.Router.getControllers()

        templateName = str(self.Router.getMethodPrefix()) + str(self.Router.getAction()) + '.html'

        return templateName

    def error_page_html(self, code):
        return host + DS + 'views' + DS + str(code) + ".html"

    def loadmodule(self):

        path = [os.path.dirname(__file__) + str("/") + str("Functions"), str(host) + str("/") + "model"]
        listpath = [path[0], path[1]]

        i = 0
        lclass = {}

        for pl in listpath:

            current_dir = os.path.join(pl)

            current_module_name = os.path.splitext(os.path.basename(current_dir))[0]

            for file in glob.glob(os.path.join(current_dir + "/*.py")):
                name = os.path.splitext(os.path.basename(file))[0]

                i = i + 1
                # Ignore __ files
                ++i
                if name.startswith("__init__"):
                    continue
                if name != "__init__":
                    lclass0 = {name: name}
                    lclass.update(lclass0)

        if Version.PYVERSION_MA <= 2:
            item = lclass.iteritems()
        else:
            item = lclass.items()
        result = {}
        for key, value in item:
            if value not in result.values():
                result[key] = value

        return result



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###
from pytonik import Log
log_msg = Log.Log()

try:
    import cx_Oracle
except Exception as err:
    log_msg.critical(err)


class Oracle:

    def __init__(self, setting):
        self.settings = setting
        self.host = setting.get('host', '') # 
        self.database = setting.get('database', '') # 
        self.username = setting.get('username', '') # 
        self.password = setting.get('password', '') # 
        self.prefix = setting.get('prefix', '') # 
        self.port = setting.get('port', '') # 
        self.Exception = ""
        self.conn =  None
        self.con = None
        self.result = None
        self.connectDB()

    def connectDB(self):

        try:
            dsn = cx_Oracle.makedsn(self.host, self.port, service_name=self.database)
            self.conn = cx_Oracle.connect(self.username, 
                                          self.password, 
                                          dsn,
                                          encoding="UTF-8")
          
        except cx_Oracle.IntegrityError as err:
            log_msg.error(err)
            self.Exception = err

    def query(self, sql="", value = ""):
        try:
            self.con = self.conn.cursor()
            if sql !="" and value != "":
                self.con.execute(str(sql), value)
            else:
                self.con.execute(str(sql))
        except Exception as err:
            log_msg.error(err)
            self.Exception = err
            
        return self

    def querymultiple(self, sql="", value = ""):
        try:
            self.con = self.conn.cursor()
            if sql !="" and value != "":
                self.con.executemany(str(sql), value)
            else:
                self.con.executemany(str(sql))
        except Exception as err:
            log_msg.error(err)
            self.Exception = err
        return self


    def lastId(self):
        return self.con.lastrowid

    def fetch(self):
        column = [d[0] for d in self.con.description]
        def row(*args):
            return dict(zip(column, args))
        return row

    def queryone(self, sql="", value = ""):
        self.con = self.conn.cursor()
        if sql !="" and value != "":
            self.con.execute(str(sql), value)
        else:
            self.con.execute(str(sql))

        return self.con

    def all(self):
        self.result = self.fetch()
        return self.fetch()

    def one(self):
        return self.con.fetchone()

    def count(self):
        return self.con.rowcount

    def countall(self):
        self.all()
        return self.con.rowcount

    def save(self):
        try:
            self.conn.commit()
            return True
        except Exception as err:
            self.Exception = err
            log_msg.error(err)
            return self

    def close(self):
        return self.con.close()

    def create(self, TABLES = ''):
        self.con = self.conn.cursor()
        if TABLES:
            for table_name in TABLES:
                table_description = TABLES[table_name]
                try:
                    self.con.execute(table_description)
                    self.Exception = "Database table '{}' created successfully.".format(table_name)
                except cx_Oracle.IntegrityError as err:
                        log_msg.info("Database table '{}' already exists.".format(table_name))
                        self.Exception =  "Database table '{}' already exists.".format(table_name)         
            return self
        else:
            self.Exception =  "Empty Table"
            return self




###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###



import sys, os, re
from pytonik import Version, Log
from pytonik.Config import Config
from pytonik.Core.env import env
from pytonik.Core import Helpers
from pytonik.Controllers import Controllers
from pytonik.util.Variable import Variable

log_msg = Log.Log()


class Router(env):
    def args(self, to="", params=[]):
        return to, params

    def __getattr__(self, item):
        return item

    def __init__(self):

        self.control = Controllers()
        self._route = []
        self._params = {}
        self._getcontrol = []
        self._getaction = []
        self._geturi = []
        self._redirect = []
        self._method = []
        self.route_to = []
        self._despatch = []
        self._code = []
        return None

    def get(self, uri, call=""):
        self._route_(route=uri, call=call, method="GET")
        return self

    def post(self, uri, call=""):
        self._route_(route=uri, call=call, method="POST")
        return self

    def put(self, uri, call=""):
        self._route_(route=uri, call=call, method="POST")
        return self

    def any(self, uri, call=""):
        self._route_(route=uri, call=call)
        return self

    def redirect(self, uri, to="", code=302):
        if to == "/":
            replace = self.control.default_controllers
        else:
            replace = to
        self._despatch.append(uri), self._redirect.append(replace), self._code.append(code)
        return self

    def permanentRedirect(self, uri, to="", code=301):
        if to == "/":
            replace = self.control.default_controllers
        else:
            replace = to

        self._despatch.append(uri), self._redirect.append(replace), self._code.append(code)
        return self


    def where(self, *args):

        if len(params) > 0:

            if isinstance(args[0], dict):
                if Version.PYVERSION_MA >= 3:

                    params = args[0].items()
                else:
                    params = args[0].iteritems()

                for k, v in params:
                    regex = re.compile(v)
                    if regex.match(self._params.get(k, "")):
                        return True
                    else:
                        return False

            elif isinstance(args, list):

                regex = re.compile(args[1])
                if regex.match(self._params.get(args[0], "")):
                    return True
                else:
                    return False

        else:
            return False


    def _route_(self, route="", call="", method=""):

        route = route.split('/')
        new_para = route

        parameter = {}
        replace = ""
        new_uri = []
        list_params = []
        params = ""

        new_paraf = []
        if self.control._getLanguages() in new_para:
            new_paraf = self.control._getUri()[1:]
            new_paraf.pop(-1)
        else:
            new_paraf = self.control._getUri()[2:]

        if self.control._getControllers() in new_paraf:
            new_paraf = self.control._getUri()
            new_paraf.pop(-1)

        if len(call) > 0:
            replace = call[0]
            params = call[1]

        if replace in route:

            new_uri = new_paraf

            for i, para in enumerate(params):

                if (len(new_uri) - i) > 0:
                    v_para = new_uri[i]
                else:
                    v_para = ""
                list_params.append(para)
                list_params.append(v_para)

            parameter = Helpers.covert_list_dict(list_params)
            self._params.update(parameter)

        if len(route) == 1:
            ctl = route[0] if route[0] != "" else self.control.default_controllers
            self._getcontrol.append(ctl)

            self._getaction.append(self.control.default_actions)
        elif len(route) > 1:
            self._getcontrol.append(route[0] if route[0] != "" else self.control.default_controllers)
            self._getaction.append(route[1] if route[1] != "" else self.control.default_actions)

        self._route.append(replace)
        self._method.append(method)

    def getParams(self):
        return self._params

    def getController(self):
        return self._getcontrol

    def getAction(self):
        return self._getaction

    def getRouter(self):
        return self._route

    def getMethod(self):
        return self._method

    def getRediret(self):
        return self._redirect

    def getDespatch(self):
        return self._despatch

    def getCode(self):
        return self._code



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
#############################################
#############################################
# Support Table Artisan Conducts Database querys
# Using Schema Pattern to controller callable funtions and methods
# Each method represent query builder attribute


from pytonik import App, Version, Log

app = App.App()


class Table:

    def __init__(self, table=""):
        self.DB = app.DB()
        self.prefix = self.DB.prefix
        self.table = str(self.prefix) + str(table)
        self.tabledict = table
        self.result = None
        self.rowCount = None
        self.table_from = ""
        self.table_select = ""
        self.table_exist = ""
        self.table_notexist = ""
        self.table_where = []
        self.table_wherenotin = ""
        self.table_wherein = ""
        self.table_orwhere = []
        self.table_drop = ""
        self.table_delete = ""
        self.table_value = ""
        self.table_groupby = ""
        self.table_orderby = ""
        self.table_offset = ""
        self.table_limit = ""
        self.table_join = []
        self.table_leftjoin = []
        self.table_rightjoin = []
        self.table_outerjoin = []
        self.table_whereBetween = ""
        self.table_whereNotBetween = ""
        self.table_max = ""
        self.table_min = ""
        self.table_count = ""
        self.table_distinct = ""
        self.table_avg = ""
        self.error = ""
        self.table_On = ""
        self.table_skip = ""
        self.table_take = ""
        self.table_having = ""
        self.table_whereisnull = []
        self.table_wherenotnull = []
        self.table_orOn = []

        return None

    def drop(self):
        self.table_drop = "DROP TABLE {exists} {table}".format(table=self.table, exists='IF ' + str(
            self.table_exist) if self.table_exist is not "" else "")
        t_result = self.DB.query(self.table_drop)
        self.error = t_result.Exception
        return self

    def exists(self, rawquery=""):
        rawString = "WHERE EXISTS ({rawquery})".format(rawquery=rawquery) if rawquery is not "" else "EXISTS"
        self.table_exist = "{rawquery}".format(rawquery=rawString)
        return self

    def notExist(self, rawquery=""):
        rawString = "WHERE NOT EXISTS ({rawquery})".format(rawquery=rawquery) if rawquery is not "" else "NOT EXISTS"
        self.table_notexist = "{rawquery}".format(rawquery=rawString)
        return self


    def select(self, *value):

        if self.table_value is not "":
            values = self.table_value

        elif len(value) > 0:

            values = ""

        else:
            values = '*'

        if self.table_where is not "":
            _and = " AND "

            _where = ""
        else:
            _and = ""
            _where = " WHERE "

        if len(self.table_orwhere) > 0:
            _or = " OR " if _where is not "WHERE" else ''

        else:
            _or = ""


        value = value if value is not "" else ""

        self.table_select = "SELECT {value}{distinct}{values}{max}{min}{count}{avg} FROM {table}{exists}{notexist}{outerjoin}{join}{leftjoin}{rightjoin}{where}{whereisnull}{wherenotnull}{whereBetween}{whereNotBetween}{wherenotin}{orwhere}{groupBy}{having}{orderBy}{limit}{take}".format(
            distinct=self.table_distinct,
            value=','.join(value) if self.table_count is "" else '',
            values=values if self.table_count is "" is "" else '',
            max=self.table_max,
            min=self.table_min,
            count=self.table_count if len(self.table_outerjoin) < 1 else '',
            avg=self.table_avg,
            table=self.table,
            exists=self.table_exist,
            whereBetween=str(" WHERE") + self.table_whereBetween if len(self.table_whereBetween) > 0 else '',
            whereNotBetween=str(" WHERE") + self.table_whereNotBetween if len(self.table_whereNotBetween) > 0 else '',
            notexist=self.table_notexist,
            outerjoin=" ".join(self.table_outerjoin),
            join=" ".join(self.table_join),
            leftjoin=" ".join(self.table_leftjoin),
            rightjoin=" ".join(self.table_rightjoin),
            whereisnull=str(" WHERE") + str(''.join(self.table_whereisnull)) if len(self.table_whereisnull) > 0 else '',
            wherenotnull=str(" WHERE") + str(''.join(self.table_wherenotnull)) if len(self.table_wherenotnull) > 0 else '',
            where=str(" WHERE") + str(_and.join(self.table_where)) if len(self.table_where) > 0 else '',
            wherenotin=str(_where) + str(_and) + str(self.table_wherenotin) if len(self.table_wherenotin) > 0 is not "" else '',
            orwhere=str(_or) + str(_or.join(self.table_orwhere)),
            groupBy=self.table_groupby,
            orderBy=self.table_orderby,
            limit=self.table_limit,
            take= self.table_take,
            having= self.table_having
        )

        return self

    def union(self, string):
        self.table_select = "{} UNION {}".format(string, self.table_select)
        return self

    def unionall(self, string):
        self.table_select = "{} UNION ALL {}".format(string, self.table_select)
        return self

    def max(self, column):
        self.table_max = ",MAX({column})".format(column=column)
        return self

    def min(self, column):
        self.table_min = ",MIN({column})".format(column=column)
        return self

    def count(self, *column):
        if len(column) > 1:
            variables = column[0] if column[0] is not "" else '*'
            columnas = column[1]
            sign = "as"
        else:
            variables = column[0] if column[0] is not "" else '*'
            sign = ""
            columnas = ""

        self.table_count = "COUNT({column}) {sign} {string} ".format(column=variables, sign=sign, string=columnas)
        return self

    def avg(self, column):
        self.table_avg = ", AVG({column})".format(column=column)
        return self

    def distinct(self, column=""):
        self.table_distinct = "DISTINCT {}".format(column=column)
        return self

    def whereAnd(self, variable="", sign="", string=""):
        string_type = "'{string}'".format(string=string) if (type(string) == str) == True else string
        self.table_distinct = "AND {variable} {sign} {string}".format(variable=variable, sign=sign, string=string_type)
        return self

    def whereIn(self, variable, list=[]):

        self.table_wherenotin = "{variable} NOT IN {list}".format(variable=variable, list='({})'.format(",".join(list)))

        return self

    def whereNotIn(self, variable, list=[]):

        self.table_wherenotin = "{variable} NOT IN {list}".format(variable=variable, list='({})'.format(",".join(list)))

        return self

    def orOn(self, variable):
        if len(variable) > 2:
            variables = variable[0]
            sign = variable[1]
            string = variable[2]
        else:
            variables = variable[0]
            sign = '='
            string = variable[1]
        table_orOn = " OR {variable} {sign} {string}".format(variable=variables, sign=sign, string=string)

        self.table_orOn.append(table_orOn)
        return self

    def on(self, variable):

        if len(variable) > 2:

            variables = variable[0]
            sign = variable[1]
            string = variable[2]

        else:
            variables = variable[0]
            sign = '='
            string = variable[1]

        self.table_On = " ON {variable} {sign} {string}".format(variable=variables, sign=sign, string=string)

        return self

    def orWhere(self, *variable):
        if len(variable) > 2:
            variables = variable[0]
            sign = variable[1]
            string = variable[2]
        else:
            variables = variable[0]
            sign = '='
            string = variable[1]

        string_type = "'{string}'".format(string=string) if (type(string) == str) == True else string

        table_orwhere = "{variable} {sign} {string}".format(variable=variables, sign=sign, string=string_type)
        self.table_orwhere.append(table_orwhere)

        return self

    def whereColumn(self, *variable):

        if len(variable) > 2:
            variables = variable[0]
            sign = variable[1]
            string = variable[2]
        else:
            variables = variable[0]
            sign = '='
            string = variable[1]

        if (type(variables) == tuple) == True:
            lv = []
            for v in variable:
                lv.append(" ".join(v))
            self.table_select = "SELECT FROM {table} WHERE {variable} ".format(table=self.table,
                                                                               variable=" AND ".join(lv))

        elif (type(variables) == str) == True:

            self.table_select = "SELECT FROM {table} WHERE {variable} {sign} {string} ".format(table=self.table,
                                                                                               variable=variables,
                                                                                               sign=sign, string=string)

        return self

    def whereBetween(self, column, values=[]):

        self.table_whereBetween = " {column} BETWEEN {values} ".format(column=column, values=" AND ".join(values))

        return self

    def whereNotBetween(self, column, values=[]):

        self.table_whereNotBetween = " {column} NOT BETWEEN {values} ".format(column=column,
                                                                              values=" AND ".join(values))

        return self

    def whereBetweenRaw(self, column, values=[]):

        column = " {column} BETWEEN {values} ".format(column=column, values=" AND ".join(values))

        return column

    def whereRaw(self, string):
        rawset = string
        return rawset

    def selectRaw(self, string):
        rawset = string
        return rawset

    def where(self, *variable):
        if len(variable) > 2:
            variables = variable[0]
            sign = variable[1]
            string = variable[2]
        else:
            variables = variable[0]
            sign = '='
            string = variable[1]

        string_type = "'{string}'".format(string=string) if (type(string) == str) == True else string

        table_where = " {variable} {sign} {string}".format(variable=variables, sign=sign, string=string_type)

        self.table_where.append(table_where)
        return self

    def whereNull(self, string):

        self.table_whereisnull = " {string} IS NULL".format(string=string)
        return self



    def whereNotNull(self, string):
        self.table_wherenotnull = "  {string} IS NOT NULL".format(string=string)
        return self

    def delete(self):
        self.table_delete = "DELETE FROM {table} {where}".format(table=self.table, where=self.table_where)
        t_result = self.DB.query(self.table_delete)
        return t_result.save() if t_result.Exception == "" else t_result.Exception

    def groupBy(self, *values):
        self.table_groupby = " GROUP BY {values}".format(values=' , '.join(values))
        return self

    def having(self, *values):
        self.table_having = " HAVING {values}".format(values=' , '.join(values))
        return self

    def havingRaw(self, values):
        havingRaw = " HAVING {values}".format(values=' , '.join(values))
        return havingRaw

    def orHavingRaw(self, values):
        orHavingRaw = " HAVING {values}".format(values=' OR '.join(values))

        return orHavingRaw


    def orderBy(self, value, sort=""):

        self.table_orderby = " ORDER BY {value} {sort}".format(value=value, sort=sort)

        return self

    def offset(self, offset=""):
        self.table_offset = str(offset) + ',' if offset is not "" else 0
        return self

    def limit(self, limits):

        self.table_limit = " LIMIT {offset} {limit}".format(offset=self.table_offset, limit=limits)
        return self

    def skip(self, offset=""):
        self.table_skip = str(offset) + ',' if offset is not "" else 0

        return self

    def take(self, limits):

        self.table_take = " LIMIT {offset} {limit}".format(offset=self.table_offset, limit=limits)
        return self


    def value(self, *values):
        self.table_value = "{values}".format(values=' , '.join(values))
        return self

    def join(self, *variable):

        if len(variable) > 3:
            table = variable[0]
            variables = variable[1]
            sign = variable[2]
            string = variable[3]
        elif len(variable) == 3:
            table = variable[0]
            variables = variable[1]
            sign = '='
            string = variable[2]
        else:

            table = variable[0]
            variables = variable[1]
            sign = ""
            string = ""

        table_join = " JOIN {table} ON {variable} {sign} {string}".format(table=str(self.prefix) + str(table),
                                                                          variable=variables, sign=sign, string=string)
        self.table_join.append(table_join)

        return self

    def outerJoin(self, *variable):

        if len(variable) > 3:
            table = str(variable[0])
            variables = variable[1]
            sign = variable[2]
            string = variable[3]
        elif len(variable) == 3:
            table = str(variable[0])
            variables = variable[1]
            sign = '='
            string = variable[2]
        else:

            table = variable[0]
            variables = variable[1]
            sign = ""
            string = ""


        table_outerjoin = " LEFT OUTER JOIN  {table} ON {variable} {sign} {string}".format(table=table,
                                                                                           variable=variables,
                                                                                           sign=sign, string=string)
        self.table_outerjoin.append(table_outerjoin)
        return self

    def fromTable(self, *table):

        self.table_from = "FROM {}".format("AS".join(table))
        return self

    def leftJoin(self, *variable):

        if len(variable) > 3:
            table = variable[0]
            variables = variable[1]
            sign = variable[2]
            string = variable[3]
        elif len(variable) == 3:
            table = variable[0]
            variables = variable[1]
            sign = '='
            string = variable[2]
        else:

            table = variable[0]
            variables = variable[1]
            sign = ""
            string = ""

        table_leftjoin = " LEFT JOIN {table} ON {variable} {sign} {string}".format(table=str(self.prefix) + str(table),
                                                                                   variable=variables, sign=sign,
                                                                                   string=string)

        self.table_leftjoin.append(table_leftjoin)

        return self

    def rightJoin(self, *variable):
        if len(variable) > 3:
            table = variable[0]
            variables = variable[1]
            sign = variable[2]
            string = variable[3]
        elif len(variable) == 3:
            table = variable[0]
            variables = variable[1]
            sign = '='
            string = variable[2]
        else:

            table = variable[0]
            variables = variable[1]
            sign = ""
            string = ""

        table_rightjoin = " RIGHT JOIN {table} ON {variable} {sign} {string}".format(
            table=str(self.prefix) + str(table), variable=variables, sign=sign, string=string)
        self.table_rightjoin.append(table_rightjoin)
        return self

    def pluck(self, column, key=""):

        self.table_select = "SELECT {column} FROM {table} ".format(table=str(self.table), column="*" if len(column) < 1 else ' {}.{}'.format(str(self.table), column))
        self.get()
        kl = []
        if key is "":
            for v in self.result:
                if Version.PYVERSION_MA <= 2:
                    lt = v.iteritems()
                else:
                    lt = v.items()
                for ks, vs in lt:
                    ls = vs
                kl.append(ls)
        else:
            for v in self.result:
                kd = {}
                if Version.PYVERSION_MA <= 2:
                    lt = v.iteritems()
                else:
                    lt = v.items()
                for ks, vs in lt:
                    l = {key: vs}
                    kd.update({key: vs})

                kl.append(kd)

        return kl

    def chunk(self, number, funcquery=""):

        if self.table_value is not "":
            values = self.table_value

        else:
            values = '*'

        if self.table_where is not "":
            _and = " AND "

            _where = ""
        else:
            _and = ""
            _where = " WHERE "

        if len(self.table_orwhere) > 0:
            _or = " OR " if _where is not "WHERE" else ''

        else:
            _or = ""



        self.table_select = "SELECT {distinct}{values}{max}{min}{count}{avg} FROM {table}{exists}{notexist}{outerjoin}{join}{leftjoin}{rightjoin}{where}{whereisnull}{wherenotnull}{whereBetween}{whereNotBetween}{wherenotin}{orwhere}{groupBy}{having}{orderBy}{limit}{take}".format(
            distinct=self.table_distinct,

            values=values if self.table_count is "" is "" else '',
            max=self.table_max,
            min=self.table_min,
            count=self.table_count if len(self.table_outerjoin) < 1 else '',
            avg=self.table_avg,
            table=self.table,
            exists=self.table_exist,
            whereBetween=str(" WHERE") + self.table_whereBetween if len(self.table_whereBetween) > 0 else '',
            whereNotBetween=str(" WHERE") + self.table_whereNotBetween if len(self.table_whereNotBetween) > 0 else '',
            notexist=self.table_notexist,
            outerjoin=" ".join(self.table_outerjoin),
            join=" ".join(self.table_join),
            leftjoin=" ".join(self.table_leftjoin),
            rightjoin=" ".join(self.table_rightjoin),
            whereisnull=str(" WHERE") + str(''.join(self.table_whereisnull)) if len(self.table_whereisnull) > 0 else '',
            wherenotnull=str(" WHERE") + str(''.join(self.table_wherenotnull)) if len(self.table_wherenotnull) > 0 else '',
            where=str(" WHERE") + str(_and.join(self.table_where)) if len(self.table_where) > 0 else '',
            wherenotin=str(_where) + str(_and) + str(self.table_wherenotin) if len(self.table_wherenotin) > 0 is not "" else '',
            orwhere=str(_or) + str(_or.join(self.table_orwhere)),
            groupBy=self.table_groupby,
            orderBy=self.table_orderby,
            limit=self.table_limit,
            take= self.table_take,
            having= self.table_having)

        self.get()
        num = int(number)

        kl = []
        if funcquery is "":

            for v in self.result:
                num -= 1
                kd = {}
                for ks, vs in v.items():
                    if num > 0:
                        kd.update({ks: vs})
                ++num
                kl.append(kd)

            return kl
        else:
            result = ""
            for v in self.result:
                kd = {}
                for ks, vs in v.items():

                    if num > 0:
                        lf = '{}'.format(ks)
                        if '{'+lf+'}' in funcquery:

                            table_update = str(funcquery).replace('{'+lf+'}', str(vs))

                            t_result = self.DB.query(table_update)

                            result = t_result.save() if t_result.Exception == "" else t_result.Exception
            return result



    def find(self, num=0):
        self.table_find = "SELECT * FROM {table} ".format(table=str(self.table))
        t_result = self.DB.query(self.table_find)
        r = ""
        if t_result.Exception == "":
            get = self.get()
            if get.rowCount > 0:
                rg = get.result
                for l in rg:
                    rf = {}
                    if Version.PYVERSION_MA <= 2:
                        lt = l.iteritems()
                    else:
                        lt = l.items()
                    for k, v in lt:
                        lk = l[k]
                        if lk == num:
                            r = l
            else:
                r = ""
        else:
            r = t_result.Exception

        return r

    def update(self, data=[]):
        if (type(data) == list):
            if len(data) > 0:
                value = []
                column = []
                for l in data:
                    value.append(l)
                    if Version.PYVERSION_MA <= 2:
                        lt = l.iteritems()
                    else:
                        lt = l.items()
                    for k, v in lt:
                        if k not in column:
                            column.append('{k}="{v}"'.format(k=k, v=v))

                lcolumn = ' , '.join(column)

                table_update = "UPDATE {table} SET {column} {where}".format(table=str(self.table),
                                                                            column=lcolumn, where=self.table_where)
                t_result = self.DB.query(table_update)
                return t_result.save() if t_result.Exception == "" else t_result.Exception
            else:
                return "Empty Data"
        else:
            return "Only Accepts type list"

    def insert(self, data=[]):
        if (type(data) == list) is True:
            if len(data) > 0:
                ksys = []
                value = []
                val = []
                column = []
                il = 1
                for l in data:
                    value.append(l)
                    if Version.PYVERSION_MA <= 2:
                        lt = l.iteritems()
                    else:
                        lt = l.items()
                    il += 1
                    ksys.append(':{}'.format(il))
                    for k, v in lt:
                        if k not in column:
                            column.append(k)
                    val.append(tuple(l.values()))

                lcolumn = ' , '.join(column)
                kvariables = ' ,'.join(ksys)

                table_insert = "INSERT INTO  {table}  ({column}) VALUES ({kvariables}) ".format(
                    table=str(self.table), column=lcolumn, kvariables=kvariables)

                if len(value) == 1:
                    t_result = self.DB.query(table_insert, val[0])
                else:
                    t_result = self.DB.querymultiple(table_insert, val)
                return t_result.save() if t_result.Exception == "" else t_result.Exception
            else:
                return "Empty Data"
        else:
            return "Only Accepts type list"

    def insertGetId(self, data=[]):
        if (type(data) == list) is True:
            if len(data) > 0:
                ksys = []
                value = []
                val = []
                column = []
                il = 1
                for l in data:
                    value.append(l)
                    if Version.PYVERSION_MA <= 2:
                        lt = l.iteritems()
                    else:
                        lt = l.items()
                    il += 1
                    ksys.append(':{}'.format(il))
                    for k, v in lt:
                        if k not in column:
                            column.append(k)
                    val.append(tuple(l.values()))

                lcolumn = ' , '.join(column)
                kvariables = ' ,'.join(ksys)

                table_insert = "INSERT INTO  {table}  ({column}) VALUES ({kvariables}) ".format(
                    table=str(self.table), column=lcolumn, kvariables=kvariables)

                if len(value) == 1:
                    t_result = self.DB.query(table_insert, val[0])
                else:
                    t_result = self.DB.querymultiple(table_insert, val)
                return t_result.lastId() if t_result.Exception == "" else t_result.Exception
            else:
                return "Empty Data"
        else:
            return "Only Accepts type list"

    def create(self):
        if type(self.tabledict) == dict:
            for table_name in self.tabledict:
                table_description = self.tabledict[table_name]
                t_result = self.DB.query(table_description)
            return t_result.save() if t_result.Exception == "" else t_result.Exception
        else:
            return "Accepts type dictionary"

    def raw(self, rawstring):
        return rawstring

    def clear(self):
        if Version.PYVERSION_MA <= 2 and Version <= 7:
            self.table_where[:]
            self.table_orwhere[:]
            self.table_whereisnull[:]
            self.table_wherenotnull[:]
            self.table_join[:]
            self.table_leftjoin[:]
            self.table_rightjoin[:]
            self.table_outerjoin[:]
        elif Version.PYVERSION_MA == 3 and Version.PYVERSION_MI <= 3:
            self.table_where[:]
            self.table_orwhere[:]
            self.table_whereisnull[:]
            self.table_wherenotnull[:]
            self.table_join[:]
            self.table_leftjoin[:]
            self.table_rightjoin[:]
            self.table_outerjoin[:]
        elif Version.PYVERSION_MA == 3 and Version.PYVERSION_MI >= 4:
            self.table_where.clear()
            self.table_orwhere.clear()
            self.table_join.clear()
            self.table_leftjoin.clear()
            self.table_rightjoin.clear()
            self.table_outerjoin.clear()
            self.table_whereisnull.clear()
            self.table_wherenotnull.clear()

        self.table_select = ""
        self.table_whereBetween = ""
        self.table_whereNotBetween = ""
        self.table_from = ""
        self.table_exist = ""
        self.table_notexist = ""
        self.table_wherenotin = ""
        self.table_wherein = ""
        self.table_drop = ""
        self.table_delete = ""
        self.table_value = ""
        self.table_groupby = ""
        self.table_orderby = ""
        self.table_offset = ""
        self.table_limit = ""
        self.table_whereBetween = ""
        self.table_whereNotBetween = ""
        self.table_max = ""
        self.table_min = ""
        self.table_count = ""
        self.table_distinct = ""
        self.table_avg = ""
        self.error = ""
        self.table_On = ""
        self.table_skip = ""
        self.table_take = ""
        self.table_having = ""

    def first(self):
        self.get()
        return self.result[0]

    def get(self):
        t_result = self.DB.query(self.table_select)

        self.error = t_result.Exception
        self.result = t_result.fetch()
        self.rowCount = t_result.count()
        self.clear()

        return self

    def set(self):
        return self.table_select


# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 2/24/20.

class Exception:

    def __getattr__(self, item):
        return item

    def __init__(self, error=""):
         self.error = error
         
    def __str__(self):
        return self.error




# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 19/11/2019.

from datetime import datetime, date, timedelta


class now:

    def __getattr__(self, item):
        return item

    def __init__(self, *args, **kwargs):

        self.fmt = ""
        self.date_time = ""
        if len(args) > 0 or len(kwargs) > 0:

            if all(args) != False:
                self.ag_o = self.datetime(*args, **kwargs)
            else:
                self.ag_o = self.datetime(**kwargs)

        return None

    def __str__(self):
        return self.ag_o


    def ago(self, string="", format=""):

        self.date_time = string

        self.fmt = self.format(format)
        delta = self.delta()
        if delta.find(',') > 0:
            days, hours = delta.split(',')
            days = int(days.split()[0].strip())
            hours, minutes = hours.split(':')[0:2]
        else:
            hours, minutes = delta.split(':')[0:2]
            days = 0
        days, hours, minutes = int(days), int(hours), int(minutes)
        datelets = []
        years, months, xdays = None, None, None
        plural = lambda x: 's' if x != 1 else ''
        if days >= 365:
            years = int(days / 365)
            datelets.append('%d year%s' % (years, plural(years)))
            days = days % 365
        if days >= 30 and days < 365:
            months = int(days / 30)
            datelets.append('%d month%s' % (months, plural(months)))
            days = days % 30
        if not years and days > 0 and days < 30:
            xdays = days
            datelets.append('%d day%s' % (xdays, plural(xdays)))
        if not (months or years) and hours != 0:
            datelets.append('%d hour%s' % (hours, plural(hours)))
        if not (xdays or months or years):
            datelets.append('%d minute%s' % (minutes, plural(minutes)))

        return ', '.join(datelets) + ' ago.'




    def delta(self):
        current_datetime = datetime.now()

        try:
            return str(current_datetime - self.fmt)
        except Exception as err:
            return err


    def format(self, fmt=''):

        fmt = '%Y-%m-%d %H:%M:%S' if fmt == "" else  fmt
        try:
            return datetime.strptime(self.date_time, fmt)
        except Exception as err:
            return err

    def time(self, format='%H:%M:%S'):
        try:
            return datetime.strftime(datetime.now(), format)

        except Exception as err:
            return err


    def date(self, format="%Y-%m-%d"):
        try:
            return datetime.strftime(datetime.now(), format)

        except Exception as err:
            return err

    def datetime(self, format="%Y-%m-%d %H:%M:%S"):

        try:
            return datetime.strftime(datetime.now(), format)

        except Exception as err:
            return err

    def create(self, string="", oldformat="%Y-%m-%d %H:%M:%S", newformat="%d-%m-%Y"):

        try:

            dt = datetime.strptime(str(string), str(oldformat))
            return dt.strftime(str(newformat))

        except Exception as err:
            return err

    def timestamp(self, string = ""):

        try:
            date_s = datetime.fromtimestamp(int(string)) if string !="" else datetime.now().strftime("%s")

            return date_s

        except Exception as err:
            return err

    def past(self, days=0,  seconds=0,minutes=0,hours=0, weeks=0):
        try:
            return  datetime.now() - timedelta(days=int(days), seconds=int(seconds),minutes=int(minutes), hours=int(hours), weeks=int(weeks))
        except Exception as err:
            return err


    def future(self, days=0,  seconds=0,minutes=0,hours=0, weeks=0):
        try:
            return  datetime.now() + timedelta(days=int(days), seconds=int(seconds),minutes=int(minutes), hours=int(hours), weeks=int(weeks))
        except Exception as err:
            return err


    def subtract(self, date1='', format1='',  date2='', format2=''):

        try:
            d1 = datetime.strptime(date1, format1)

            d2 = datetime.strptime(date2, format2)
            return abs((d2 - d1).days)
        except Exception as err:
            return err










# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 09/11/2019.


from pytonik.Editor import HTMLeditor
from pytonik.Log import Log
from pytonik.App import App
import os
log_msg = Log()
Ap = App()

class public():

    def __getattr__(self, item):
        return item
    
    def __init__(self, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:
            if all(args) is not False:
                self.pt = self.path(*args, **kwargs)
            else:
                self.pt = self.path(**kwargs)

        return None

    def __str__(self):

        return self.pt


    def path(self, public = ""):
        DS = str('/');

        if public == "/":
            DS = ""
        else:
            DS = "/"

        return DS + 'public' + DS + value




###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


def App():
    from . import App
    return App.App()


def Version():
    from . import Version as vern
    return vern


def Config():
    from . import Config as Config
    return Config.Config()


def Helpers():
    from .Core import Helpers
    return Helpers


def Session():
    from .Session import Session
    return Session()


def Request():
    from .Request import Request
    return Request()


def Hash():
    from .Hash import Hash
    return Hash()


def Model():
    from .Model import Model
    M = Model()
    return M


def Load(m):
    M = Model()
    return M.load(m)


def Functions():
    from .Core import Functions
    return Functions


def Pagination():
    from pytonik.Functions.pagination import pagination
    return pagination


def Validation():
    from pytonik.Functions.validation import validation
    return validation


def curl():
    from pytonik.Functions.curl import curl
    return curl


def now():
    from pytonik.Functions.now import now
    return now


def os():
    from pytonik.Functions.agent import os
    return os


def browser():
    from pytonik.Functions.agent import browser
    return browser


def extend():
    from pytonik.Functions.extend import extend
    return extend


def path():
    from pytonik.Functions.path import path
    return path


def include():
    from pytonik.Functions.include import include
    return include


def File():
    from .Core import File

    return File


def SendEmail(from_send, to_recipient, message_subject="", messege_content=""):
    from .Core.SMTP import SMTP

    return SMTP().send(from_send, to_recipient, message_subject, messege_content)


def Logs():
    from .Log import Log

    return Log()


def url(path="", lang=False):
    from pytonik.Router import Router
    import os
    ront = Router()
    dev_path = ""
    http_s = os.environ.get("HTTP_HOST")
    urlv = os.environ.get('REQUEST_URI', os.environ.get('PATH_INFO'))
    if http_s == "127.0.0.1" or http_s == "localhost":
        dev_path = str("/") + str(urlv).split('/')[1]
    else:
        dev_path = ""

    seturl = str("localhost:") + str(os.environ.get("SERVER_PORT", ''))
    http = os.environ.get("HTTPS")

    if http == 'on':
        url = str("https://") + os.environ.get("HTTP_HOST", seturl) + str(dev_path) + "/" + ront.alllanguages.get(
            ront.getLanguages(), ront.getLanguages()) if lang is True else str("https://") + os.environ.get("HTTP_HOST",
                                                                                                            seturl) + str(
            dev_path)
    else:
        url = str("http://") + os.environ.get("HTTP_HOST", seturl) + str(dev_path) + "/" + ront.alllanguages.get(
            ront.getLanguages(), ront.getLanguages()) if lang is True else str("http://") + os.environ.get("HTTP_HOST",
                                                                                                           seturl) + str(
            dev_path)

    DS = ""
    p = ""

    if path == "":
        DS = ""
    else:
        if path[:1] == "/":
            p = path[1:]
            DS = "/"
        else:
            p = path
            DS = "/"

    return url + DS + p



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 07/12/2019.
import os
from pytonik import Version

if Version.PYVERSION_MA < 3 and  Version.PYVERSION_MI <= 7:
    from urllib2 import urlopen

else:
    from urllib.request import urlopen
import json

class ip:

    def __init__(self):

        self.HTTP_headers = ['SERVER_ADDR',
                            'REMOTE_ADDR',
                            'HTTP_CLIENT_IP'
                            ]
        self.HTTP_proxy_header = ['HTTP_VIA',
                                  'VIA',
                                  'Proxy-Connection',
                                  'HTTP_X_FORWARDED_FOR',
                                  'HTTP_FORWARDED_FOR',
                                  'HTTP_X_FORWARDED',
                                  'HTTP_FORWARDED',
                                  'HTTP_CLIENT_IP',
                                  'HTTP_FORWARDED_FOR_IP',
                                  'X-PROXY-ID',
                                  'MT-PROXY-ID',
                                  'X-TINYPROXY',
                                  'X_FORWARDED_FOR',
                                  'FORWARDED_FOR',
                                  'X_FORWARDED',
                                  'FORWARDED',
                                  'CLIENT-IP',
                                  'CLIENT_IP',
                                  'PROXY-AGENT',
                                  'HTTP_X_CLUSTER_CLIENT_IP',
                                  'FORWARDED_FOR_IP',
                                  'HTTP_PROXY_CONNECTION'
                                  ]
        self.ips = ['127.0.0.0', '127.0.0.1', '127.0.0.2', '192.0.0.0', '192.0.0.1', '192.168.0.0',
               '192.168.0.1', '192.168.0.253', '192.168.0.254', '192.168.0.255', '192.168.1.0',
               '192.168.1.1', '192.168.1.253', '192.168.1.254', '192.168.1.255', '192.168.2.0',
               '192.168.2.1', '192.168.2.253', '192.168.2.254', '192.168.2.255', '10.0.0.0', '10.0.0.1',
               '11.0.0.0', '11.0.0.1', '1.0.0.0', '1.0.1.0', '1.1.1.1', '255.0.0.0', '255.0.0.1',
               '255.255.255.0', '255.255.255.254', '255.255.255.255', '0.0.0.0', '::', '0::', '::1',
               '0:0:0:0:0:0:0:0']


        self.ip = ""
        self.hostname = ""
        self.city = ""
        self.region = ""
        self.country = ""
        self.loc = ""
        self.org = ""
        self.is_vpn = False
        return None

    def get(self):

        for header in self.HTTP_headers:
            if self.os(header) is not "":
                self.ip = '127.0.0.0' if self.os(header) in self.ips else self.os(header)
                self.property(self.ip)
            return self



    def os(self, header):
        return os.environ.get(header, '')

    def check(self, ip):
        for header in self.HTTP_proxy_header:
            response = True if self.os(header) is not "" else False
        return response

    def vpn(self):

        for header in self.HTTP_proxy_header:
            if self.os(header) is not "":
                self.ip = self.os(header)
                self.is_vpn = self.check(self, self.ip)
                self.property(self.ip)
            return self


    def property(self, ip):
        try:
            ipU= "http://ipinfo.io/{ip}/json".format(ip=ip)
            url_ip = urlopen(ipU).read()
            response = url_ip.decode('utf-8')
            j_response = json.loads(response)
            if j_response.get("bogon", "") == "":
                self.hostname = j_response.get("hostname", "")
                self.city = j_response.get("city", "")
                self.region = j_response.get("region", "")
                self.country = j_response.get("country", "")
                self.loc = j_response.get("loc", "")
                self.org = j_response.get("org", "")
            return self
        except Exception:
            return False



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


import os
import datetime
import sys
import ast
from pytonik import Version
from pytonik.util.Variable import Variable

try:
    from http import cookies as cook
except Exception as err:
    import Cookie as cook




class Session(Variable):

    def __getattr__(self, item):
        return item

    def __call__(self, *args, **kwargs):
        return None

    def __init__(self):
        self.result = ""
        self.s_string = "HTTP_COOKIE"
        self.session_list = []
        return None
        
    def has(self, key=""):

        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
            session_dict = self.x_get()
            if len(session_dict) > 0:
                if session_dict.get(key, "") != "":
                    return True
                elif session_dict.get(' {key}'.format(key=key), "") != "":
                    return True
                else:
                    return False
            else:
                return False
        else:

            if self.get(key) != "" and self.get(key) != None:

                return True
            else:
                return False


    def set(self, key="", value="", duration=3600, url="", path="/"):
        url_v = self.out("HTTP_HOST")+str(":")+str(self.out("SERVER_PORT", '')) if self.out("HTTP_HOST") == "localhost" or self.out("HTTP_HOST") == "127.0.0.1" else self.out("HTTP_HOST")
        url = url if url  != "" else url_v
        expires = datetime.datetime.utcnow() + datetime.timedelta(minutes=duration)  # minutes in 30 days

        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
            session_string = str(key)+"="+str(value)
            self.set_x(session_string) 
        else:

            cooKeys = cook.SimpleCookie(self.out(self.s_string))

            cooKeys[str(key)] = value
            cooKeys[str(key)]['domain'] = url
            cooKeys[str(key)]['path'] = '/'
            cooKeys[str(key)]['expires'] = expires.strftime('%a, %d %b %Y %H:%M:%S')

            print(cooKeys)



        
    def get(self, key=""):
        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:
            session_dict = self.x_get()
            if len(session_dict) > 0:
                if session_dict.get(key, "") != None or session_dict.get(key, "") != "":
                    try:
                        return ast.literal_eval(session_dict.get(key, ""))
                    except Exception as err:
                        return session_dict.get(key, "")
                elif session_dict.get(' {key}'.format(key=key), "") != None or session_dict.get(' {key}'.format(key=key), "") != "":
                    try:
                        return ast.literal_eval(session_dict.get(' {key}'.format(key=key), ""))
                    except Exception as err:
                        return session_dict.get(' {key}'.format(key=key), "")
                else:
                    return ""
            else:
                return ""       
        else:
            cooKeys = cook.SimpleCookie()
            OsEnviron = self.out(self.s_string)
            
            if OsEnviron != None:
                cooKeys.load(OsEnviron)
                if key in cooKeys:

                    if cooKeys[key].value != None or cooKeys[key].value != "":

                        try:
                            return ast.literal_eval(cooKeys[key].value)
                        except Exception as err:
                            return cooKeys[key].value

                    else:
                        return ""
                else:
                    return ""
            else:
                return ""

    def destroy(self, *args):


        if self.out("SERVER_SOFTWARE") == Version.AUTHOR:

            session_result = self._delete(args)
            if session_result[0] == True:
                return self._reset(session_result[1])
            else:
                return False
            
        else:
            cooKeys = cook.SimpleCookie(self.out(self.s_string))
            if Version.PYVERSION_MA >= 3:

                cookv = cooKeys.items()
            else:
                cookv = cooKeys.iteritems()

            OsEnviron = self.out(self.s_string)

            if self.s_string in os.environ:
                if args:

                    for key in args:

                        if self.get(key) != "":
                            return self.set(key, "", 60)
                else:
                    for key, v in cookv:

                        if self.get(key) != "":
                            return self.set(key, "", 60)

            else:
                return False
    
    def set_x(self, session_string=""):
        if session_string  != "":
            if self.s_string not in self.see():
                self.default(self.s_string, session_string)
                
            else:
                self.session_list.append(session_string)
                self._update(self.unqiue(self.session_list))
        else:
            return False
                
                
    def _update(self, session_string):
        dict_session = dict({"HTTP_COOKIE": session_string})
        try:
            self.update(dict_session)
            return True
        except Exception as e:
            return False
        

    def unqiue(self, session_list=[]):
       
        session_relist = []
        for l_session in session_list:
            session_string = str(l_session)+";"+str(self.out("HTTP_COOKIE", ""))
            session_relist.append(session_string)
        initial_session = str(session_relist).replace("[' ", "").replace("]", "").replace(',', ";").replace("'", "").replace("[", "")
        return self._unquie(initial_session)

    def _unquie(self, initial_session=""):
        try:
            unique_session = initial_session.split(";")
            re_unique = []
            [ress.append(x) for x in unique_session if x not in re_unique]
            return ";".join(re_unique)
        except Exception as err:
            return initial_session    
        
    def _delete(self, session_key=tuple()):


        get_session = self.x_get()
        response = None
        if len(get_session) > 0: 
            if len(session_key) > 0:
                
                try:
                    for k in session_key:  
                        get_session.pop(k, None)
                    self._update(get_session)
                    response = True

                except Exception as err:
                    response = False 
            else:   
                try:
                    for k in dict(get_session):
                        get_session.pop(k, None)
                    response = True 
                except Exception as err:
                    response = False 
        else:
            response = False 
        return response, get_session

    def _reset(self, session_dict = dict()):

        session_list =  []
        if Version.PYVERSION_MA >= 3:
            session_dict_l = session_dict.items()
        else:
            session_dict_l = cooKsession_dicteys.iteritems()

        for k, v in session_dict_l:
            session_list.append("{k}={v}".format(k=k,v=v))
            
        return self._update(";".join(session_list))

    def x_get(self):

        session_dict = {}
        session_string = ""
        try:
            session_string = self.out(self.s_string)
        except Exception as err:

            session_dict = {}
        try:
            for c_g in session_string.split(";"):
                k, v = c_g.split("=")
                session_dict.update({str(k):str(v)})
        except Exception as err:
            try:
                k, v = session_string.split("=")
                session_dict.update({str(k):str(v)})
            except Exception as err:
                 session_dict = {}
        return session_dict



class form:
    
    def __getattr__(self, item):
            return item

    def __init__(self):
        return None
    
    def _FLOAT(self):
        return None
    
    def _STRING(self):
        return ""
    
    def _INT(self):
        return ""
    
    def _CHAR(self):
        return ""
        
    
        


from pytonik.Web import App, Version

app = App()


def index():
    app.header()
    # do work here


def page400():
    data = {
        'title': '400 NOT Found!',
        'name': 'Pytonik',
        'version': Version().VERSION_TEXT

    }
    return app.views('400', data)


def page404():
    data = {
        'title': '404 NOT Found!',
        'name': 'Pytonik',
        'version': Version().VERSION_TEXT
    }
    return app.views('405', data)


def page405():
    data = {
        'title': '405 NOT Found!',
        'name': 'Pytonik',
        'version': Version().VERSION_TEXT
    }
    return app.views('405', data)



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 05/11/2019.
import os

from pytonik.Functions.url import url
from pytonik.util.Variable import Variable
from pytonik import Version


class path(url, Variable):
    def __getattr__(self, item):
        return item

    def __call__(self, *args, **kwargs):

        return None

    def __init__(self, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:
            if all(args) is not False:
                self.pt = self.path(*args, **kwargs)
            else:
                self.pt = self.path(**kwargs)

        return None

    def __str__(self):

        return self.pt

    def path(self, path="", link=False):

        DS = str('/')
        u = ""

        dev_path = ""

        if path[:1] == DS or path[0] == DS:
            DS = ""
        else:
            DS = "/"

        if bool(link) == True:
            u = self.url()

        return str(u)+str(DS)+str(path)

    def exist(self, newpath, defaultpath="", link=False):

        if os.path.isfile(newpath) == True:

            return self.path(newpath, link)

        elif os.path.isfile(os.getcwd() + newpath) == True:
            return self.path(newpath, link)

        elif os.path.isfile(self.public(newpath)) == True:
            return self.path(self.public(newpath), link)

        elif os.path.isdir(newpath) == True:
            return self.path(newpath, link)


        elif os.path.isdir(self.public(newpath)) == True:

            return self.path(self.public(newpath), link)

        else:

            if defaultpath != "":

                if os.path.isfile(defaultpath) == True:

                    return self.path(defaultpath, link)

                elif os.path.isfile(os.getcwd() + defaultpath) == True:
                    return self.path(defaultpath, link)

                elif os.path.isfile(self.public(defaultpath)) == True:

                    return self.path(self.public(defaultpath), link)

                elif os.path.isdir(defaultpath) == True:
                    return self.path(defaultpath, link)

                elif os.path.isdir(self.public(defaultpath)) == True:

                    return self.path(self.public(defaultpath), link)
            else:
                return False

    def public(self, path):
        
        host = str(os.path.dirname(os.getcwd())) + '/public'
        
        DS = str('/')

        if path[:1] == DS or path[:1] == DS:
            DS = ""
        else:
            DS = "/"

        return str(host) + DS + str(path)



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 2/24/20.


class env:



    def __init__(self):
        return None

    def _e(self):

        from pytonik import Version, Config


        import sys, os

        if os.path.isdir(os.getcwd() + '/public'):
            host = os.getcwd()  # os.path.dirname(os.getcwd())

        else:
            host = os.path.dirname(os.getcwd())

        DS = str("/")

        envpath = host + DS + ".env"

        if os.path.isfile(envpath) == True:

            try:
                f = open(envpath, "r")
                return f.read()
            except Exception as err:
                   print(err)
        else:

            print(".env file not found")



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

from pytonik.App import App
from pytonik import Log
from pytonik.Driver.Schema import Schema
import sys, os, importlib
log_msg = Log.Log()



class Model(Schema):
    
    def __init__(self):
        ap = App()
        self.db = ap.DB()




    def load(self, m):

        if os.path.isdir(os.getcwd() + '/public'):
            host = os.getcwd()  # os.path.dirname(os.getcwd())

        else:
            host = os.path.dirname(os.getcwd())

        DS = str("/")


        paths = host + DS + 'model'
        model = paths + DS + m + ".py"
        sys.path.append(paths)
        importlib._RELOADING
        if os.path.isfile(model) == True:

            try:

                md = importlib.import_module(m)
                ob = getattr(md, m)
                if hasattr(ob(), '__call__'):
                    return ob()
                else:
                    log_msg.error("'%s' is not a callable" % m)

            except Exception as err:
                log_msg.error(err)
                self.App.header(0)

        else:

            log_msg.error("Model {e} does not exist ".format(m))

    


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

import os
from pytonik import Log
from pytonik import Version
from pytonik.Controllers import Controllers
from pytonik import Router
import ast

log_msg = Log.Log()


class Lang:

    def __init__(self, lg):
        self.data = ""
        self.lg = lg


    def loadLang(self):
        if os.path.isdir(os.getcwd() + '/public'):
            host = os.getcwd()  # os.path.dirname(os.getcwd())

        else:
            host = os.path.dirname(os.getcwd())

        DS = str("/")
        ront = Controllers()
        getl = ront.all_languages

        langpath = host + DS + 'lang'+DS+getl.get(self.lg.lower(), self.lg.lower()) +".py"

        try:
            if os.path.isfile(langpath) == True:

                with open(langpath, 'rb') as rb:
                    self.data = rb.read().decode('utf-8')
                return self.data

        except Exception as e:
            log_msg.error("Lang file not found {}".format(e))
            return "Lang file not found {}".format(e)



    def get(self, key, defindValue = ''):

        data = ast.literal_eval(self.data)

        return data.get(key.lower(), defindValue)





# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 2/25/20.
##pip install --install-option="--prefix=$PREFIX_PATH" package_name



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 1/19/20.


import argparse
import locale
import os
import re
import sys
import time
import warnings
from pytonik import Version
from pytonik.Core.requirement import requirement
from typing import Any, Callable, Dict, List, Pattern, Union
from random import randint
from typing import Any, Callable, Dict, List, Pattern, Union
from pytonik.cmd.console import (  # type: ignore
    colorize, bold, red, green, turquoise, nocolor, color_terminal
)
from random import randint

try:
    import readline

    if readline.__doc__ and 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
        USE_LIBEDIT = True
    else:
        readline.parse_and_bind("tab: complete")
        USE_LIBEDIT = False
except ImportError:
    USE_LIBEDIT = False



class ValidationError(Exception):
    """Raised for validation errors."""


def is_path(x: str) -> str:
    x = os.path.expanduser(x)
    if not os.path.isdir(x):
        raise ValidationError(__("Please enter a valid path name."))
    return x


def boolean(x: str) -> bool:
    if x.upper() not in ('Y', 'YES', 'N', 'NO'):
        raise ValidationError(__("Please enter either 'y' or 'n'."))
    return x.upper() in ('Y', 'YES')


def allow_empty(x: str) -> str:
    return x


def nonempty(x: str) -> str:
    if not x:
        raise ValidationError(__("Please enter some text."))
    return x


def __(mes_id):

    try:
        userlang = locale.getlocale()[0]

        l_ = mes_id
        for l in lang.lang:
            if l == str(userlang):
                getla = lang.lang.get(l, '')

                if getla != "":
                    l_ =  getla.get(mes_id, '')
                    if l_ != "":
                        l_ = getla.get(mes_id, '')
                    else:
                        l_ = mes_id
                else:
                    l_ = mes_id
    except Exception as arr:
        l_ = mes_id
    return l_

def doTraceBack():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
    traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    traceback.print_exc()



if sys.platform == 'win32':
    # On Windows, show questions as bold because of color scheme of PowerShell (refs: #5294).
    COLOR_QUESTION = 'bold'
else:
    COLOR_QUESTION = 'purple'

PROMPT_PREFIX = '> '


def do_prompt(text: str, default: str = None, validator: Callable[[str], Any] = nonempty) -> Union[str, bool]:  # NOQA

    while True:
        if default is not None:
            prompt = PROMPT_PREFIX + '%s [%s]: ' % (text, default)
        else:
            prompt = PROMPT_PREFIX + text + ': '
        if USE_LIBEDIT:
            # Note: libedit has a problem for combination of ``input()`` and escape
            # sequence (see #5335).  To avoid the problem, all prompts are not colored
            # on libedit.
            pass
        else:
            prompt = colorize(COLOR_QUESTION, prompt, input_mode=True)
        x = term_input(prompt).strip()
        if default and not x:
            x = default
        try:
            x = validator(x)
        except ValidationError as err:
            print(red('* ' + str(err)))
            continue
        break
    return x


def ask(d: Dict) -> None:

    print(bold(__('Welcome to the Pytonik MVC Framework %s.' % Version.VERSION_TEXT)))

    print(__(bold(red('PLEASE NOTE :'))+' This feature is use for installation of project requirments.'))

    d['step'] = do_prompt(__('Do you want install requirements  (y/n)'),'n', boolean)
    if d.get('step', '') == True:
        try:
            requirement().rf()
        except Exception as err:
            print(err)

    if d.get('step', '') is False:
        sys.exit(-1)




def term_input(prompt: str) -> str:
    if sys.platform == 'win32':

        print(prompt, end='')
        return input('')
    else:
        return input(prompt)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage='%(prog)s [OPTIONS] <PROJECT_DIR>',
        epilog=__("For more information, visit < https://pytonik.readthedocs.io >."),
        description=__("""Pytonik is a python framework built to enhance web development
        fast and easy, also help web developers to build more apps with less codes.
        it uses expressive architectural pattern, structured on model view controller MVC
        and bundles of component to reuse while deploying the framework."""))

    parser.add_argument('-q', '--quit', action='store_true', dest='quit',
                        default=None,
                        help=__('quit mode'))
    parser.add_argument('--version', action='version', dest='show_version',
                        version='%%(prog)s %s' % Version.VERSION_TEXT)

    parser.add_argument('path', metavar='PROJECT_DIR', default='.', nargs='?',
                        help=__('project root'))

    return parser



def main(argv: List[str] = sys.argv[1:]) -> int:
    # parse options
    parser = get_parser()
    try:
        args = parser.parse_args(argv)

    except SystemExit as err:
        return err.code
    d = vars(args)
    ask(d)
    #path = str(os.getcwd())



if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###
from pytonik.Core.env import env
from pytonik.Config import Config

import logging, os, datetime, sys

class Log():
    def __init__(self, base_path=__name__):
        if os.path.isdir(os.getcwd() + '/public'):
            host = os.getcwd()  # os.path.dirname(os.getcwd())

        else:
            host = os.path.dirname(os.getcwd())

        self.format = '%(asctime)s, %(msecs)d %(name)s - [%(funcName)s %(levelname)s]  -    %(message)s' #[%(filename)s:%(lineno)d]
        self.datefmt='%m/%d/%Y %I:%M:%S %p'
        self.locate = host
        #base_path = str(self.locate.split('/')[-1:][0])
        #print(base_path)
        self.logs = logging.getLogger(base_path) #base_path
        dt = datetime
        self.indata = ""




    def debug(self, string=""):
        self.indata = string
        self.__file()
        if int(self.__check()) == 1:
            self.__open_tem()
            self.logs.debug(self.indata)
            exit()
        else:
            self.logs.debug(self.indata)




    def info(self, string=""):
        self.indata = string
        self.__file()
        if int(self.__check()) == 1:
            self.__open_tem()
            self.logs.info(self.indata)
            exit()
        else:
            self.logs.info(self.indata)



    def warning(self, string=""):
        self.indata = string
        self.__file()
        if int(self.__check()) == 1:
            self.__open_tem()
            self.logs.warning(self.indata)
            exit()
        else:
            self.logs.warning(self.indata)




    def error(self, string=""):
        self.indata = string
        self.__file()

        if int(self.__check()) == 1:
            self.__open_tem()
            self.logs.error(self.indata)
            exit()
        else:
            self.logs.error(self.indata)




    def critical(self, string=""):
        self.indata = string
        self.__file()
        if int(self.__check()) == 1:
            self.__open_tem()
            self.logs.critical(self.indata)
            exit()
        else:
            self.logs.critical(self.indata)



    def __file(self):
        from pytonik.util.Exception import Exception
        error_log = self.locate + "/" + 'app.log'
        if os.path.isfile(error_log) == False:
            try:
                f = open(error_log, 'a+')
                f.write("")
                f.close()
            except Exception as err:
                if conf.get('exception', "") == 1:
                    Exception(err)

        else:

            if int(self.__check()) == 1:
                logging.basicConfig(stream=sys.stdout, format=self.format, datefmt=self.datefmt)
                
            else:
                logging.basicConfig(filename=error_log, format=self.format, datefmt=self.datefmt)

    def __check(self):
        getenv = env()
        conf = Config()
        conf.add(getenv._e())
        return 1 if conf.get('exception', 0) == 1  else 0


    def __open_tem(self):
        from pytonik.App import App
        App.header()



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 08/01/2020.
import re

class iteration:


    def __getattr__(self, item):

        return item

    def __call__(self, *args, **kwargs):

        return None

    def __init__(self, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:

            if all(args) is not False:
                self.iter = self.iteri(*args, **kwargs)
            else:
                self.iter = self.iteri(**kwargs)
        return None

    def __str__(self):
        return self.iter


    def iteri(self, dictionary="", itr="pid"):

        i = 0
        if dictionary is not "" or dictionary is not None:
            dist, apend = [], []
            for l in dictionary:
                i += 1
                ++i
                listv = l
                dist = {itr: i}
                dist.update(listv)
                apend.append(dist)
            return apend



    def keyword(self, keywords, limit = 0):

        if type(keywords) == str:
            pattern = re.compile("\s*,\s*|\s+$")
            tag = pattern.split()
            splittag = [x for x in tag if x]
        elif type(keywords) == list:
            tag = keywords
            joinl = [x for x in tag if x]
            stag = " ".join(joinl)
            splittag = stag.split(",")



        try:

            if int(limit) > 0:
                lk = []
                ct = 0
                for nsl in splittag:
                    ct +=1
                    if ct < limit+1:
                        lk.append(nsl)

                return lk
            else:
                return splittag

        except Exception as err:
            return  err

